/* $Id: globus_replica_catalog.c,v 1.75 2002/04/18 16:57:38 smartin Exp $ */


#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * @file
 * 
 * Globus Replica Catalog C API Source File
 *
 * This file defines functions (and a variable) that manage the information
 * in a replica catalog.
 *
 * @see globus_replica_catalog.h
 */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

#include "globus_replica_catalog.h"
#include "globus_i_replica_catalog.h"
#include "version.h"
#include "sasl.h"

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL

/**
 * Module activation callback.
 */
static
int
globus_l_replica_catalog_activate(
    void )
{
    int                                 rc;

    rc = globus_module_activate(GLOBUS_THREAD_MODULE);
    if (rc != GLOBUS_SUCCESS) 
    {
	return rc;
    }
    rc = globus_module_activate(GLOBUS_COMMON_MODULE);
    if (rc != GLOBUS_SUCCESS) 
    {
	return rc;
    }

    return GLOBUS_SUCCESS;
}

/**
 * Module deactivation callback.
 */
static
int
globus_l_replica_catalog_deactivate(
    void )
{
    int                                 rc;

    rc = globus_module_deactivate(GLOBUS_THREAD_MODULE);
    if (rc != GLOBUS_SUCCESS) 
    {
	return rc;
    }
    rc = globus_module_deactivate(GLOBUS_COMMON_MODULE);
    if (rc != GLOBUS_SUCCESS) 
    {
	return rc;
    }

    return GLOBUS_SUCCESS;
}

/**
 * Module descriptor.
 */
globus_module_descriptor_t globus_i_replica_catalog_module = {
    "globus_replica_catalog",
    globus_l_replica_catalog_activate,
    globus_l_replica_catalog_deactivate,
    GLOBUS_NULL,
    GLOBUS_NULL,
    &local_version
};

/**
 * Merge two lists of filenames, eliminating duplicates.  Resulting
 * list is a null-terminated list of strings and is set to GLOBUS_NULL
 * if the its length is zero.  Result must be freed by the caller with
 * globus_libc_free().
 */
static
globus_result_t
globus_l_replica_catalog_merge_filenames(
    char **                             filenames_a,
    char **                             filenames_b,
    char ***                            filenames )
{
    int                                 len_a = 0;
    int                                 len_b = 0;
    int					len;
    int                                 i;
    int                                 j;

    if (filenames_a != GLOBUS_NULL)
    {
	while (filenames_a[len_a]) len_a++;
    }

    if (filenames_b != GLOBUS_NULL)
    {
	while (filenames_b[len_b]) len_b++;
    }

    if ((len_a + len_b) == 0)
    {
	*filenames = GLOBUS_NULL;
	return GLOBUS_SUCCESS;
    }

    *filenames = (char **) globus_libc_malloc( (len_a + len_b + 1)
						* sizeof(char *) );

    if (filenames_a != GLOBUS_NULL)
    {
	for (i = 0; filenames_a[i]; i++) 
	{
	    (*filenames)[i] = filenames_a[i];
	}
	(*filenames)[i] = NULL;
    }

    len = len_a;
    
    if (filenames_b != GLOBUS_NULL)
    {
	for (i = 0; filenames_b[i]; i++) 
	{
	    for (j = 0; j < len_a; j++) 
	    {
		if (strcmp(filenames_b[i], filenames_a[j]) == 0) 
		{
		    break;
		}
	    }
	    if (j == len_a) 
	    {
		(*filenames)[len++] = filenames_b[i];
	    }
	}
    }

    (*filenames)[len] = NULL;

    return GLOBUS_SUCCESS;
}

/**
 * Construct the DN of a location given its logical name and the
 * collection information in the given handle.  Result must be freed by
 * the caller with globus_libc_free().
 */
static
char *
globus_l_replica_catalog_location_dn(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name )
{
    char *                              dn;

    dn = (char *) globus_libc_malloc(
	strlen(GLOBUS_REPLICA_CATALOG_LOCATION_RDN_TYPE)
	+ strlen(location_name)
	+ strlen(handle->dn)
	+ 4 );

    sprintf( dn, "%s=%s, %s",
	     GLOBUS_REPLICA_CATALOG_LOCATION_RDN_TYPE, location_name,
	     handle->dn ); 

    return dn;
}

/**
 * Construct the DN of a logical file given its logical name and the
 * collection information in the given handle.  Result must be freed by
 * the caller with globus_libc_free().
 */
static
char *
globus_l_replica_catalog_logicalfile_dn(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          logicalfile_name )
{
    char *                              dn;

    dn = (char *) globus_libc_malloc(
	strlen(GLOBUS_REPLICA_CATALOG_LOGICALFILE_RDN_TYPE)
	+ strlen(logicalfile_name)
	+ strlen(GLOBUS_REPLICA_CATALOG_FILEPARENT_RDN)
	+ strlen(handle->dn)
	+ 6 );

    sprintf( dn, "%s=%s, %s, %s",
	     GLOBUS_REPLICA_CATALOG_LOGICALFILE_RDN_TYPE, logicalfile_name,
	     GLOBUS_REPLICA_CATALOG_FILEPARENT_RDN,
	     handle->dn );
    
    return dn;
}

/**
 * Construct the DN of a fileparent given its logical name and the
 * collection information in the given handle.  Result must be freed by
 * the caller with globus_libc_free().
 */
static
char *
globus_l_replica_catalog_fileparent_dn(
    globus_replica_catalog_collection_handle_t *    handle ) 
{
    char *                              dn;

    dn = (char *) globus_libc_malloc(
	strlen(GLOBUS_REPLICA_CATALOG_FILEPARENT_RDN)
	+ strlen(handle->dn)
	+ 3 );
    
    sprintf( dn, "%s, %s",
	     GLOBUS_REPLICA_CATALOG_FILEPARENT_RDN,
	     handle->dn );

    return dn;
}

/**
 * Create a fileparent entry in the replica catalog.
 */
static
globus_result_t
globus_l_replica_catalog_fileparent_create(
    globus_replica_catalog_collection_handle_t *    handle )
{
    char *                              fileparent_dn;
    char *                              objectclass_values[4];
    LDAPMod                             objectclass_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    objectclass_values[0] = "top";
    objectclass_values[1] = "GlobusTop";
    objectclass_values[2] = GLOBUS_REPLICA_CATALOG_FILEPARENT_OBJECTCLASS;
    objectclass_values[3] = NULL;
    objectclass_mod.mod_op = 0;
    objectclass_mod.mod_type = "objectclass";
    objectclass_mod.mod_values = objectclass_values;
    mods[0] = &objectclass_mod;
    mods[1] = NULL;
    
    fileparent_dn = globus_l_replica_catalog_fileparent_dn(handle);
    
    lderr = ldap_add_s(handle->ld, fileparent_dn, mods);
    
    globus_libc_free(fileparent_dn);

    if (lderr != LDAP_SUCCESS) 
    {
	return GLOBUS_REPLICA_CATALOG_ERROR( 
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

/**
 * Delete a fileparent entry in the replica catalog.
 */
static
globus_result_t
globus_l_replica_catalog_fileparent_delete(
    globus_replica_catalog_collection_handle_t *    handle )
{
    char *                              dn;
    int                                 lderr;
    char *                              lderrstr;

    dn = globus_l_replica_catalog_fileparent_dn(handle);

    lderr = ldap_delete_s(handle->ld, dn);
    
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS) 
    {
	if (lderr == LDAP_NOT_ALLOWED_ON_NONLEAF) 
	{
	    lderrstr = "Children entries exist";
	} 
	else 
	{
	    lderrstr = ldap_err2string(lderr);
	}
	
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    lderrstr );
    }
    
    return GLOBUS_SUCCESS;
}

#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

/* Connection Management */

globus_result_t
globus_replica_catalog_collection_handleattr_init(
    globus_replica_catalog_collection_handleattr_t *    handleattrs )
{
    globus_mutex_init(&handleattrs->mutex, GLOBUS_NULL);

    globus_mutex_lock(&handleattrs->mutex);
    {
	handleattrs->authentication_mode =
	    GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_SECURE;
	handleattrs->username = NULL;
	handleattrs->password = NULL;
    }
    globus_mutex_unlock(&handleattrs->mutex);
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_handleattr_destroy(
    globus_replica_catalog_collection_handleattr_t *    handleattrs )
{
    if (handleattrs != GLOBUS_NULL)
    {
	globus_mutex_lock(&handleattrs->mutex);
	{
	    if (handleattrs->authentication_mode ==
		GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_CLEARTEXT) 
	    { 
		globus_libc_free(handleattrs->username);
		handleattrs->username = NULL;
		globus_libc_free(handleattrs->password);
		handleattrs->password = NULL;
	    }
	    
	    handleattrs->authentication_mode =
		    GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_SECURE;
	}
	globus_mutex_unlock(&handleattrs->mutex);

	globus_mutex_destroy(&handleattrs->mutex);
    }
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_handleattr_set_authentication_mode(
    globus_replica_catalog_collection_handleattr_t *    handleattrs,
    globus_replica_catalog_authentication_mode_t        authentication_mode,
    ... )
{
    va_list                             ap;

    globus_mutex_lock(&handleattrs->mutex);
    {
	switch (authentication_mode) 
	{
	    
	case GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_SECURE:
	    if (handleattrs->authentication_mode ==
		GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_CLEARTEXT) 
	    { 
		globus_libc_free(handleattrs->username);
		handleattrs->username = NULL;
		globus_libc_free(handleattrs->password);
		handleattrs->password = NULL;
	    }
	    break;
	    
	case GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_CLEARTEXT:
	    va_start(ap, authentication_mode);
	    handleattrs->username = globus_libc_strdup(va_arg(ap, char *));
	    handleattrs->password = globus_libc_strdup(va_arg(ap, char *));
	    va_end(ap);
	    break;
	    
	default:
	    globus_mutex_unlock(&handleattrs->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"Unrecognized authentication mode" );
	}
    
	handleattrs->authentication_mode = authentication_mode;

    }
    globus_mutex_unlock(&handleattrs->mutex);
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_handleattr_get_authentication_mode(
    globus_replica_catalog_collection_handleattr_t *    handleattrs,
    globus_replica_catalog_authentication_mode_t *      authentication_mode,
    ... )
{
    va_list                             ap;
    char **                             username;
    char **                             password;
    
    globus_mutex_lock(&handleattrs->mutex);
    {
	*authentication_mode = handleattrs->authentication_mode;
	
	if (*authentication_mode ==
	    GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_CLEARTEXT) 
	{
	    va_start(ap, authentication_mode);
	    username = va_arg(ap, char **);
	    *username = handleattrs->username;
	    password = va_arg(ap, char **);
	    *password = handleattrs->password;
	    va_end(ap);
	}
    }
    globus_mutex_unlock(&handleattrs->mutex);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_handleattr_copy(
    globus_replica_catalog_collection_handleattr_t *    dest,
    globus_replica_catalog_collection_handleattr_t *    src)
{
    globus_result_t                                 rc;
    char *                                          username;
    char *                                          password;
    globus_replica_catalog_authentication_mode_t    authentication_mode;

    rc = globus_replica_catalog_collection_handleattr_init(dest);

    if(rc != GLOBUS_SUCCESS)
    {
	goto error_exit;
    }

    rc = globus_replica_catalog_collection_handleattr_get_authentication_mode(
	src,
	&authentication_mode,
	&username,
	&password);

    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_exit;
    }

    rc = globus_replica_catalog_collection_handleattr_set_authentication_mode(
	dest,
	authentication_mode,
	username,
	password);

    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_exit;
    }
    
    return GLOBUS_SUCCESS;

 destroy_exit:
    globus_replica_catalog_collection_handleattr_destroy(dest);
 error_exit:
    return rc;
}

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/* this is VERY simplified version of the function from OpenLDAP */
int sasl_interact(LDAP *ld,unsigned flags,void *defaults,void *in) {
  sasl_interact_t *interact = in;
  while( interact->id != SASL_CB_LIST_END ) {
    const char *dflt = interact->defresult;
    interact->result = strdup( (dflt && *dflt) ? dflt : "" );
    interact->len = interact->result ? strlen( interact->result ) : 0;
    interact++;
  };
  return LDAP_SUCCESS;
}
 
/**
 * Open a connection to a catalog and initialize the connection handle.
 */
static
globus_result_t
globus_l_replica_catalog_open(
    globus_replica_catalog_collection_handle_t *        handle,
    globus_replica_catalog_collection_handleattr_t *    handleattrs,
    char *                                              url )
{
    int ldurlerr;
    char * ldurlerrstr;
    int lderr;

    globus_mutex_init(&handle->mutex, GLOBUS_NULL);

    globus_mutex_lock(&handle->mutex);
    {
	ldurlerr = ldap_url_parse(url, &(handle->urld));

	if (ldurlerr != 0) 
	{
	    globus_mutex_unlock(&handle->mutex);
	    globus_mutex_destroy(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"Can't parse LDAP URL." );
	}

	if ( (handle->urld->lud_host == NULL) || 
	     (strlen(handle->urld->lud_host) == 0) ) 
	{
	    /* there is no default host */
	    ldap_free_urldesc(handle->urld);
	    globus_mutex_unlock(&handle->mutex);
	    globus_mutex_destroy(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		"URL has no host" );
	}
    
	if (handle->urld->lud_port == 0) 
	{
	    /* there is a default port, though */
	    handle->urld->lud_port = LDAP_PORT;
	}

	if ( (handle->ld = ldap_init(
	    handle->urld->lud_host,
	    handle->urld->lud_port ) ) == NULL ) 
	{
	    ldap_free_urldesc(handle->urld);
	    globus_mutex_unlock(&handle->mutex);
	    globus_mutex_destroy(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(errno) );
	}

	/* we do this not knowing if the DN connected to exists */
	handle->dn = globus_libc_strdup(handle->urld->lud_dn);

	handle->attrs = globus_libc_malloc(
	    sizeof(globus_replica_catalog_collection_handleattr_t));
	globus_replica_catalog_collection_handleattr_copy(handle->attrs,
							  handleattrs);
	
	globus_mutex_lock(&handleattrs->mutex);
	{
          void *defaults;
          unsigned sasl_flags = LDAP_SASL_AUTOMATIC;
          char     *sasl_mech = "gsi-GSSAPI";
          struct berval passwd = { 0, NULL };
          char     *binddn = NULL;
          int      version;
          version = LDAP_VERSION3;
          if(handle->attrs->authentication_mode != 
                GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_SECURE) {
            /* always try secure authenticaton first, but hide it */
            sasl_flags = LDAP_SASL_QUIET;
          };
          ldap_set_option(handle->ld,LDAP_OPT_PROTOCOL_VERSION,&version);
          lderr = ldap_sasl_interactive_bind_s(handle->ld, binddn,
                       sasl_mech, NULL, NULL,
                       sasl_flags, sasl_interact,NULL);
          if ( lderr != LDAP_SUCCESS ) {
            /* fallback to plain username/password */
            lderr = ldap_simple_bind_s( handle->ld,
                                        handle->attrs->username,
                                        handle->attrs->password );
            if ( lderr != LDAP_SUCCESS ) {
              /* try protocol v2 */
              version = LDAP_VERSION2;
              ldap_set_option(handle->ld,LDAP_OPT_PROTOCOL_VERSION,&version);
              lderr = ldap_simple_bind_s( handle->ld,
                                          handle->attrs->username,
                                          handle->attrs->password );
            };
          };
	};
	globus_mutex_unlock(&handleattrs->mutex);
	
	if ( lderr != LDAP_SUCCESS ) 
	{
	    ldap_unbind( handle->ld );
	    ldap_free_urldesc(handle->urld);
	    globus_mutex_unlock(&handle->mutex);
	    globus_mutex_destroy(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

globus_result_t
globus_replica_catalog_collection_open(
    globus_replica_catalog_collection_handle_t *        handle,
    globus_replica_catalog_collection_handleattr_t *    handleattrs,
    char *                                              url )
{
    globus_result_t * err;
    int lderr;
    char * attributes [] = { "c", NULL }; /* dummy attrs for check */
    LDAPMessage * result;

    if ( (err = globus_l_replica_catalog_open(
	    handle,
	    handleattrs,
	    url ) ) != GLOBUS_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    globus_error_get(err),
	    "Can't open collection" );
    }

    /* test if the DN connected to exists */
    globus_mutex_lock(&handle->mutex);
    {
	if ( ( lderr = ldap_search_s( handle->ld,
		handle->dn,
		LDAP_SCOPE_BASE,
		"(objectclass=*)",
		attributes, 1,
		&result ) ) != LDAP_SUCCESS )
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
	ldap_msgfree(result);
    }
    globus_mutex_unlock(&handle->mutex);
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_list_collections(
    globus_replica_catalog_collection_handle_t * handle,
    char **                                      attribute_names,
    struct globus_replica_catalog_entry_set_s *  collections )
{
    char *                              filter = "(objectclass="
	GLOBUS_REPLICA_CATALOG_COLLECTION_OBJECTCLASS ")";
    int                                 lderr;
    
    globus_mutex_lock(&handle->mutex);
    {
	lderr = ldap_search_s(
	    handle->ld,
	    handle->dn,
	    LDAP_SCOPE_SUBTREE,
	    filter,
	    (attribute_names != GLOBUS_NULL) ? attribute_names : NULL,
	    0,
	    &(collections->result) );
    }
    globus_mutex_unlock(&handle->mutex);
	
    if ( lderr != LDAP_SUCCESS ) 
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    collections->collection = handle;

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_set_dn(
    globus_replica_catalog_collection_handle_t * handle,
    char *                                       collection_dn )
{
    globus_mutex_lock(&handle->mutex);
    {
	/* TODO: if this has been strdup'ed to before we need to free it */
	handle->dn = globus_libc_strdup(collection_dn);
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_create(
    globus_replica_catalog_collection_handle_t *     handle,
    globus_replica_catalog_collection_handleattr_t * handleattrs,
    char *                                           url,
    char **                                          filenames,
    char *                                           objectclass )
{
    globus_result_t *                   err;
    int                                 lderr;
    char *                              lderrstr;
    char *                              objectclass_values[5];
    LDAPMod                             objectclass_mod;
    LDAPMod                             filename_mod;
    LDAPMod *                           mods[3];
    
    if ( (err = globus_l_replica_catalog_open(
	    handle,
	    handleattrs,
	    url ) ) != GLOBUS_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    globus_error_get(err),
	    "Can't open collection" );
    }

    globus_mutex_lock(&handle->mutex);
    {
	objectclass_values[0] = "top";
	objectclass_values[1] = "GlobusTop";
	objectclass_values[2] = GLOBUS_REPLICA_CATALOG_COLLECTION_OBJECTCLASS;
	
	if ( (objectclass != GLOBUS_NULL)
	     && ( strcmp(objectclass,
			 GLOBUS_REPLICA_CATALOG_COLLECTION_OBJECTCLASS )
		  != 0 ) )
	{
	    objectclass_values[3] = objectclass;
	    objectclass_values[4] = NULL;
	}
	else
	{
	    objectclass_values[3] = NULL;
	}

	objectclass_mod.mod_op = 0;
	objectclass_mod.mod_type = "objectclass";
	objectclass_mod.mod_values = objectclass_values;
	mods[0] = &objectclass_mod;

	if (filenames != GLOBUS_NULL)
	{
	    filename_mod.mod_op = 0;
	    filename_mod.mod_type =
		GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE;
	    filename_mod.mod_values = filenames;
	    mods[1] = &filename_mod;
	    mods[2] = NULL;
	}
	else
	{
	    mods[1] = NULL;
	}

	if ( ( lderr = ldap_add_s( handle->ld,
				   handle->dn,
				   mods ) ) != LDAP_SUCCESS )
	{
	    if (lderr == LDAP_NO_SUCH_OBJECT) {
		lderrstr = "Parent entry doesn't exist";
	    }
	    else
	    {
		lderrstr = ldap_err2string(lderr);
	    }

	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		lderrstr );
	}
    
	/* XXX: Inconsistent State: The next
	 * globus_l_replica_catalog_fileparent_create() may fail while the
	 * previous ldap_add_s() does not.  Subsequent logical file
	 * functions should check for this.
	 */

	if ( (err = globus_l_replica_catalog_fileparent_create(handle))
	     != GLOBUS_SUCCESS )
	{
	    globus_mutex_unlock(&handle->mutex);
	    globus_replica_catalog_collection_close(handle);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		globus_error_get(err),
		"Can't create fileparent entry" );
	}
    }
    globus_mutex_unlock(&handle->mutex);
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_close(
    globus_replica_catalog_collection_handle_t *    handle )
{
    if (handle != GLOBUS_NULL)
    {
	globus_mutex_lock(&handle->mutex);
	{
	    globus_replica_catalog_collection_handleattr_destroy(handle->attrs);
	    globus_libc_free(handle->attrs);
            handle->attrs = GLOBUS_NULL;
	    ldap_unbind(handle->ld);
	    handle->ld = NULL;
	    ldap_free_urldesc(handle->urld);
	    handle->urld = NULL;
	    globus_libc_free(handle->dn);
	    handle->dn = NULL;
	}
        globus_mutex_unlock(&handle->mutex);
        globus_mutex_destroy(&handle->mutex);
    }
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_delete(
    globus_replica_catalog_collection_handle_t *    handle )
{
    globus_result_t *                   err;
    int                                 lderr;
    char *                              lderrstr;

    globus_mutex_lock(&handle->mutex);
    {
    
	if ( (err = globus_l_replica_catalog_fileparent_delete(handle))
	     != GLOBUS_SUCCESS )
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		globus_error_get(err),
		"Can't delete fileparent entry" );
	}

	/* XXX: Inconsistent State: The next ldap_delete_s() may fail while
	 * the previous globus_l_replica_catalog_fileparent_delete() does
	 * not.  Subsequent logical file functions should check for this. */

	if ( ( lderr = ldap_delete_s( handle->ld,
				      handle->dn ) ) != LDAP_SUCCESS )
	{
	    if (lderr == LDAP_NOT_ALLOWED_ON_NONLEAF)
	    {
		lderrstr = "Children entries still exist";
	    }
	    else
	    {
		lderrstr = ldap_err2string(lderr);
	    }
	    
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		lderrstr );
	}
    }
    globus_mutex_unlock(&handle->mutex);
    
    globus_replica_catalog_collection_close(handle);

    return GLOBUS_SUCCESS;
}

/* Collection Information Management */

globus_result_t
globus_replica_catalog_collection_add_filenames(
    globus_replica_catalog_collection_handle_t *    handle,
    char **                                         filenames,
    globus_bool_t                                   add )
{
    LDAPMod                             filename_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;
    char **                             existing_filenames;
    char **                             replacement_filenames;

    filename_mod.mod_op = LDAP_MOD_ADD;
    filename_mod.mod_type = GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE;
    filename_mod.mod_values = filenames;
    mods[0] = &filename_mod;
    mods[1] = NULL;

    globus_mutex_lock(&handle->mutex);
    {
    
	lderr = ldap_modify_s( handle->ld,
			       handle->dn,
			       mods );

	if ( (lderr == LDAP_TYPE_OR_VALUE_EXISTS)
	     && (add == GLOBUS_TRUE) )
	{
	    
	    /* XXX: Race Condition: Another client can change filename
	     * information associated with the collection between calls to
	     * globus_replica_catalog_collection_list_filenames() and
	     * ldap_modify(). */

            globus_replica_catalog_collection_list_filenames(
                handle,
                &existing_filenames );

	    globus_l_replica_catalog_merge_filenames(
		filenames,
		existing_filenames,
		&replacement_filenames );

	    globus_libc_free(existing_filenames);

	    if (replacement_filenames != GLOBUS_NULL)
	    {
		filename_mod.mod_op = LDAP_MOD_REPLACE;
		filename_mod.mod_values = replacement_filenames;
	    
		lderr = ldap_modify_s( handle->ld,
				       handle->dn, 
				       mods );

		globus_libc_free(replacement_filenames);

		if (lderr != LDAP_SUCCESS)
		{
		    globus_mutex_unlock(&handle->mutex);
		    return GLOBUS_REPLICA_CATALOG_ERROR(
			GLOBUS_NULL,
			ldap_err2string(lderr) );
		}
	    }
	}
	else if (lderr != LDAP_SUCCESS)
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_add_attribute(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          attribute_name,
    char *                                          attribute_value )
{
    char *                              attribute_values[2];
    LDAPMod                             attribute_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    if ( strcmp( attribute_name,
		 GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE ) == 0 )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    "Can't add filenames" );
    }

    attribute_values[0] = attribute_value;
    attribute_values[1] = NULL;
    attribute_mod.mod_op = LDAP_MOD_ADD;
    attribute_mod.mod_type = attribute_name;
    attribute_mod.mod_values = attribute_values;
    mods[0] = &attribute_mod;
    mods[1] = NULL;

    globus_mutex_lock(&handle->mutex);
    {
	lderr = ldap_modify_s( handle->ld,
			       handle->dn,
			       mods ) ;
    }
    globus_mutex_unlock(&handle->mutex);

    if ( lderr != LDAP_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_delete_filenames(
    globus_replica_catalog_collection_handle_t *    handle,
    char **                                         filenames )
{
    globus_result_t                     err;
    globus_replica_catalog_entry_set_t  locations;
    int                                 count;
    LDAPMod                             filename_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    /* XXX: Race Condition: Another client can change filename
     * information associated with a location between calls to
     * globus_replica_catalog_collection_find_locations() and
     * ldap_modify_s(). */


    
    if ( ( err = globus_replica_catalog_collection_find_locations(
               handle,
               filenames,
               GLOBUS_FALSE,
               GLOBUS_NULL,
               &locations ) ) != GLOBUS_SUCCESS )
    {
        return GLOBUS_REPLICA_CATALOG_ERROR(
            globus_error_get(err),
            "Can't find locations" );
    }
    
    if ( ( err = globus_replica_catalog_entry_set_count(
               &locations,
               &count )  ) != GLOBUS_SUCCESS )
    {
        return GLOBUS_REPLICA_CATALOG_ERROR(
            globus_error_get(err),
            "Can't count locations" );
    }
    
    if (count > 0)
    {
        return GLOBUS_REPLICA_CATALOG_ERROR(
            GLOBUS_NULL,
            "Filenames still exist in locations" );
    }
    
    /* TODO: Add checking for the existence of corresponding
     * logicalfile entries. */
    
    filename_mod.mod_op = LDAP_MOD_DELETE;
    filename_mod.mod_type = GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE;
    filename_mod.mod_values = filenames;
    mods[0] = &filename_mod;
    mods[1] = NULL;
    
    globus_mutex_lock(&handle->mutex);
    {
	if ( ( lderr = ldap_modify_s( handle->ld,
				      handle->dn,
				      mods ) ) != LDAP_SUCCESS )
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_delete_attribute(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          attribute_name,
    char *                                          attribute_value )
{
    char *                              attribute_values[2];
    LDAPMod                             attribute_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    if ( strcmp( attribute_name,
		 GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE ) == 0 )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    "Can't delete filenames" );
    }

    attribute_values[0] = attribute_value;
    attribute_values[1] = NULL;
    attribute_mod.mod_op = LDAP_MOD_DELETE;
    attribute_mod.mod_type = attribute_name;
    attribute_mod.mod_values = (attribute_value == GLOBUS_NULL)
	? NULL : attribute_values;
    mods[0] = &attribute_mod;
    mods[1] = NULL;

    globus_mutex_lock(&handle->mutex);
    {
    
	lderr = ldap_modify_s( handle->ld,
			       handle->dn,
			       mods );
    }
    globus_mutex_unlock(&handle->mutex);
    
    if ( lderr != LDAP_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_list_filenames(
    globus_replica_catalog_collection_handle_t *    handle,
    char ***                                        filenames )
{
    char *                              filter = "(objectclass=*)";
    char *                              attributes [] = {
	GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE, NULL };
    int                                 lderr;
    LDAPMessage *                       result;
    LDAPMessage *                       entry;

    
    if ( ( lderr = ldap_search_s( handle->ld,
                                  handle->dn,
                                  LDAP_SCOPE_BASE,
                                  filter,
                                  attributes,
                                  0,
                                  &result ) ) != LDAP_SUCCESS )
    {
        return GLOBUS_REPLICA_CATALOG_ERROR(
            GLOBUS_NULL,
            ldap_err2string(lderr) );
    }
	
    if ( ( entry = ldap_first_entry( handle->ld,
                                     result ) ) != NULL )
    {
        if ( (*filenames = ldap_get_values(
                  handle->ld,
                  entry,
                  GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE ) ) != NULL )
        {
            ldap_msgfree(result);
            return GLOBUS_SUCCESS;
        }
    } else {
        ldap_get_option(handle->ld, LDAP_OPT_ERROR_NUMBER, &lderr);
        if (lderr != LDAP_SUCCESS)
        {
            return GLOBUS_REPLICA_CATALOG_ERROR(
                GLOBUS_NULL,
                ldap_err2string(lderr) );
        }
    }
    
    *filenames = GLOBUS_NULL;
    
    ldap_msgfree(result);
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_list_attributes(
    globus_replica_catalog_collection_handle_t *    handle,
    char **                                         attribute_names,
    struct globus_replica_catalog_attribute_set_s * attributes )
{
    char *                              filter = "(objectclass=*)";
    LDAPMessage *                       result;
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
    
	if ( ( lderr = ldap_search_s(
	    handle->ld,
	    handle->dn,
	    LDAP_SCOPE_BASE,
	    filter,
	    (attribute_names != GLOBUS_NULL) ? attribute_names : NULL,
	    0,
	    &result ) ) != LDAP_SUCCESS )
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}

	attributes->collection = handle;
	attributes->result = result;
	attributes->entry = ldap_first_entry(handle->ld, result);

	ldap_get_option(handle->ld, LDAP_OPT_ERROR_NUMBER, &lderr);
	if (lderr != LDAP_SUCCESS)
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);
    
    return GLOBUS_SUCCESS;
}

/* Location Information Management */

globus_result_t
globus_replica_catalog_location_create(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char *                                          url,
    char **                                         filenames,
    char *                                          objectclass )
{
    char *                              dn;
    char *                              objectclass_values[5];
    char *                              url_values[2];
    LDAPMod                             objectclass_mod; 
    LDAPMod                             filename_mod;
    LDAPMod                             url_mod;
    LDAPMod *                           mods[4];
    int                                 lderr;
    char *                              lderrstr;

    globus_mutex_lock(&handle->mutex);
    {
    
	dn = globus_l_replica_catalog_location_dn(handle, location_name);
	
	objectclass_values[0] = "top";
	objectclass_values[1] = "GlobusTop";
	objectclass_values[2] = GLOBUS_REPLICA_CATALOG_LOCATION_OBJECTCLASS;
	if ( (objectclass != GLOBUS_NULL)
	     && ( strcmp(objectclass,
			 GLOBUS_REPLICA_CATALOG_LOCATION_OBJECTCLASS ) != 0 ) )
	{
	    objectclass_values[3] = objectclass;
	    objectclass_values[4] = NULL;
	}
	else
	{
	    objectclass_values[3] = NULL;
	}
	objectclass_mod.mod_op = 0;
	objectclass_mod.mod_type = "objectclass";
	objectclass_mod.mod_values = objectclass_values;
	url_values[0] = url;
	url_values[1] = NULL; 
	url_mod.mod_op = 0;
	url_mod.mod_type = "uc";
	url_mod.mod_values = url_values;
	mods[0] = &objectclass_mod;
	if (filenames != GLOBUS_NULL)
	{
	    filename_mod.mod_op = 0;
	    filename_mod.mod_type =
		GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE;
	    filename_mod.mod_values = filenames;
	    mods[1] = &filename_mod;
	    mods[2] = &url_mod;
	    mods[3] = NULL;
	}
	else
	{
	    mods[1] = &url_mod;
	    mods[2] = NULL;
	}

	lderr = ldap_add_s(handle->ld, dn, mods);
	
	globus_libc_free(dn);
    }
    globus_mutex_unlock(&handle->mutex);

    if (lderr != LDAP_SUCCESS)
    {
	if (lderr == LDAP_NO_SUCH_OBJECT)
	{
	    lderrstr = "Parent entry doesn't exist";
	}
	else
	{
	    lderrstr = ldap_err2string(lderr);
	}
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    lderrstr );
    }
    
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_location_add_filenames(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char **                                         filenames,
    globus_bool_t                                   add )
{
    char **                             collection_filenames;
    int                                 i;
    int                                 j;
    int                                 found;
    char *                              dn;
    LDAPMod                             filename_mod;
    LDAPMod *                           mods[2];
    char **                             existing_filenames;
    char **                             replacement_filenames;
    int                                 lderr;

    /* XXX: Race Condition: Another client can change filename
     * information associated with the collection between calls to
     * globus_replica_catalog_collection_list_filenames() and
     * globus_replica_catalog_location_list_filenames(). */

    globus_mutex_lock(&handle->mutex);
    {
    	globus_replica_catalog_collection_list_filenames(
	    handle,
	    &collection_filenames );

	for (i = 0; filenames[i]; i++)
	{
	    found = GLOBUS_FALSE;
	    if (collection_filenames != GLOBUS_NULL) {
		for (j = 0; collection_filenames[j]; j++)
		{
		    if (strcmp(filenames[i], collection_filenames[j]) == 0)
		    {
			found = GLOBUS_TRUE;
			break;
		    }
		}
	    }
	    if (found == GLOBUS_FALSE)
	    {
		globus_mutex_unlock(&handle->mutex);
		return GLOBUS_REPLICA_CATALOG_ERROR(
		    GLOBUS_NULL,
		    "Filenames don't exist in collection" );
	    }
	}
	globus_libc_free(collection_filenames);

	dn = globus_l_replica_catalog_location_dn(handle, location_name);

	filename_mod.mod_op = LDAP_MOD_ADD;
	filename_mod.mod_type = GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE;
	filename_mod.mod_values = filenames;
	mods[0] = &filename_mod;
	mods[1] = NULL;

	lderr = ldap_modify_s(handle->ld, dn, mods);

	if ( (lderr == LDAP_TYPE_OR_VALUE_EXISTS)
	     && (add == GLOBUS_TRUE) )
	{

	    /* XXX: Race Condition: Another client can change filename
	     * information associated with the location between calls to
	     * globus_replica_catalog_location_list_filenames() and
	     * ldap_modify_s(). */
	    
	    globus_replica_catalog_location_list_filenames(
		handle,
		location_name,
		&existing_filenames );

	    globus_l_replica_catalog_merge_filenames(
		filenames,
		existing_filenames,
		&replacement_filenames );

	    globus_libc_free(existing_filenames);

	    if (replacement_filenames != GLOBUS_NULL)
	    {
		filename_mod.mod_op = LDAP_MOD_REPLACE;
		filename_mod.mod_values = replacement_filenames;

		lderr = ldap_modify_s(handle->ld, dn, mods);
		
		globus_libc_free(replacement_filenames);
	    
		if (lderr != LDAP_SUCCESS)
		{
		    globus_libc_free(dn);
		    globus_mutex_unlock(&handle->mutex);
		    return GLOBUS_REPLICA_CATALOG_ERROR(
			GLOBUS_NULL,
			ldap_err2string(lderr) );
		}
	    }
	}
	else if (lderr != LDAP_SUCCESS)
	{
	    globus_libc_free(dn);
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);

    globus_libc_free(dn);
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_location_add_attribute(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char *                                          attribute_name,
    char *                                          attribute_value )
{
    char *                              dn;
    char *                              attribute_values[2];
    LDAPMod                             attribute_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    if ( strcmp( attribute_name,
		 GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE ) == 0 )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    "Can't add filenames" );
    }

    globus_mutex_lock(&handle->mutex);
    {
    
	dn = globus_l_replica_catalog_location_dn(handle, location_name);
	
	attribute_values[0] = attribute_value;
	attribute_values[1] = NULL;
	attribute_mod.mod_op = LDAP_MOD_ADD;
	attribute_mod.mod_type = attribute_name;
	attribute_mod.mod_values = attribute_values;
	mods[0] = &attribute_mod;
	mods[1] = NULL;
	
	lderr = ldap_modify_s(handle->ld, dn, mods);
    }
    globus_mutex_unlock(&handle->mutex);
	
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_location_delete(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name )
{
    char *                              dn;
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_location_dn(handle, location_name);
	
	lderr = ldap_delete_s(handle->ld, dn);
    }
    globus_mutex_unlock(&handle->mutex);
	
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_location_delete_filenames(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char **                                         filenames )
{
    char *                              dn;
    LDAPMod                             filename_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_location_dn(handle, location_name);
	
	filename_mod.mod_op = LDAP_MOD_DELETE;
	filename_mod.mod_type = GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE;
	filename_mod.mod_values = filenames;
	mods[0] = &filename_mod;
	mods[1] = NULL;

	lderr = ldap_modify_s(handle->ld, dn, mods);
    }
    globus_mutex_unlock(&handle->mutex);
    
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_location_delete_attribute(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char *                                          attribute_name,
    char *                                          attribute_value )
{
    char *                              dn;
    char *                              attribute_values[2];
    LDAPMod                             attribute_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    if ( strcmp( attribute_name,
		 GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE ) == 0 )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    "Can't add filenames" );
    }
    
    attribute_values[0] = attribute_value;
    attribute_values[1] = NULL;
    attribute_mod.mod_op = LDAP_MOD_DELETE;
    attribute_mod.mod_type = attribute_name;
    attribute_mod.mod_values = (attribute_value == GLOBUS_NULL)
	? NULL : attribute_values;
    mods[0] = &attribute_mod;
    mods[1] = NULL;

    globus_mutex_lock(&handle->mutex);
    {
    	dn = globus_l_replica_catalog_location_dn(handle, location_name);

	lderr = ldap_modify_s(handle->ld, dn, mods);
    }
    globus_mutex_unlock(&handle->mutex);

    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_find_locations(
    globus_replica_catalog_collection_handle_t *    handle,
    char **                                         filenames,
    globus_bool_t                                   all,
    char **                                         attribute_names,
    struct globus_replica_catalog_entry_set_s *     locations )
{
    int                                 i;
    int                                 len;
    char *                              filter;
    int                                 lderr;

    for (i = 0, len = 4; filenames[i]; i++)
    {
	len += strlen(filenames[i]) + 11;
    }

    filter = (char *) globus_libc_malloc(len + 2);
    *filter = '\0';
    strcat(filter, (all == GLOBUS_TRUE) ? "(&" : "(|" );

    for (i = 0; filenames[i]; i++)
    {
	strcat( filter, "("
		GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE "=" );
	strcat(filter, filenames[i]);
	strcat(filter, ")");
    }
    strcat(filter, ")");

    globus_mutex_lock(&handle->mutex);
    {	
	lderr = ldap_search_s(
	    handle->ld,
	    handle->dn,
	    LDAP_SCOPE_ONELEVEL,
	    filter,
	    (attribute_names != GLOBUS_NULL) ? attribute_names : NULL, 0,
	    &(locations->result) );
    }
    globus_mutex_unlock(&handle->mutex);
    
    globus_libc_free(filter);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    locations->collection = handle;

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_list_locations(
    globus_replica_catalog_collection_handle_t *    handle,
    char **                                         attribute_names,
    struct globus_replica_catalog_entry_set_s *     locations )
{
    char *                              filter = "(objectclass="
	GLOBUS_REPLICA_CATALOG_LOCATION_OBJECTCLASS ")";
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
    
	lderr = ldap_search_s(
	    handle->ld,
	    handle->dn,
	    LDAP_SCOPE_ONELEVEL,
	    filter,
	    (attribute_names != GLOBUS_NULL) ? attribute_names : NULL, 0,
	    &(locations->result));
    }
    globus_mutex_unlock(&handle->mutex);
    
    if ( lderr != LDAP_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    locations->collection = handle;

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_location_list_filenames(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char ***                                        filenames )
{
    char *                              dn;
    char *                              filter = "(objectclass=*)";
    char *                              attributes [] = {
	GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE, NULL };
    LDAPMessage *                       result;
    LDAPMessage *                       entry;
    int                                 lderr;

    dn = globus_l_replica_catalog_location_dn(handle, location_name);
    
    lderr = ldap_search_s( handle->ld,
                           dn,
                           LDAP_SCOPE_BASE,
                           filter,
                           attributes, 0,
                           &result );

    globus_libc_free(dn);
    
    if (lderr != LDAP_SUCCESS)
    {
        return GLOBUS_REPLICA_CATALOG_ERROR(
            GLOBUS_NULL,
            ldap_err2string(lderr) );
    }
    
    if ( ( entry = ldap_first_entry( handle->ld,
                                     result ) ) != NULL )
    {
        if ( (*filenames = ldap_get_values(
                  handle->ld,
                  entry,
                  GLOBUS_REPLICA_CATALOG_FILENAME_ATTRIBUTE_TYPE ) ) != NULL )
        {
            ldap_msgfree(result);
            return GLOBUS_SUCCESS;
        }
    }
    else
    {
        ldap_get_option(handle->ld, LDAP_OPT_ERROR_NUMBER, &lderr);
        if (lderr != LDAP_SUCCESS)
        {
            return GLOBUS_REPLICA_CATALOG_ERROR(
                GLOBUS_NULL,
                ldap_err2string(lderr) );
        }
    }
	
    *filenames = GLOBUS_NULL;
    
    ldap_msgfree(result);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_location_list_attributes(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char **                                         attribute_names,
    struct globus_replica_catalog_attribute_set_s * attributes )
{
    char *                              dn;
    char *                              filter = "(objectclass=*)";
    LDAPMessage *                       result;
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {	
    
	dn = globus_l_replica_catalog_location_dn(handle, location_name);
	
	lderr = ldap_search_s(
	    handle->ld,
	    dn,
	    LDAP_SCOPE_BASE,
	    filter,
	    (attribute_names != GLOBUS_NULL) ? attribute_names : NULL, 0,
	    &result );

	globus_libc_free(dn);
	
	if (lderr != LDAP_SUCCESS)
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}

	attributes->collection = handle;
	attributes->result = result;
	attributes->entry = ldap_first_entry(handle->ld, result);
	
	ldap_get_option(handle->ld, LDAP_OPT_ERROR_NUMBER, &lderr);
	if (lderr != LDAP_SUCCESS)
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

/* File Object Information Management */

globus_result_t
globus_replica_catalog_logicalfile_create(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          logicalfile_name,
    int                                             size,
    char *                                          objectclass )
{
    char *                              dn;
    char *                              objectclass_values[5];
    char *                              size_values[2];
    LDAPMod                             objectclass_mod;
    LDAPMod                             size_mod;
    LDAPMod *                           mods[3];
    int                                 lderr;
    globus_result_t *                   err;

    objectclass_values[0] = "top";
    objectclass_values[1] = "GlobusTop";
    objectclass_values[2] = GLOBUS_REPLICA_CATALOG_LOGICALFILE_OBJECTCLASS;
    if ( (objectclass != GLOBUS_NULL)
	 && ( strcmp(
	     objectclass,
	     GLOBUS_REPLICA_CATALOG_LOGICALFILE_OBJECTCLASS ) != 0 ) )
    {
	objectclass_values[3] = objectclass;
	objectclass_values[4] = NULL;
    }
    else
    {
	objectclass_values[3] = NULL;
    }
    objectclass_mod.mod_op = 0;
    objectclass_mod.mod_type = "objectclass";
    objectclass_mod.mod_values = objectclass_values;
    size_values[0] = (char *) globus_libc_malloc(11 * sizeof(char));
    size_values[1] = NULL;
    sprintf(size_values[0], "%d", size); 
    size_mod.mod_op = 0;
    size_mod.mod_type = "size";
    size_mod.mod_values = size_values;
    mods[0] = &objectclass_mod;
    mods[1] = &size_mod;
    mods[2] = NULL;

    globus_mutex_lock(&handle->mutex);
    {
    
	dn = globus_l_replica_catalog_logicalfile_dn(handle, logicalfile_name);
	
	lderr = ldap_add_s(handle->ld, dn, mods);
	
	if (lderr == LDAP_NO_SUCH_OBJECT)
	{
	    /* XXX: Inconsistent State: The fileparent entry doesn't
	     * exist so we try to create it. */
	    if ( ( err = globus_l_replica_catalog_fileparent_create(
		handle ) ) != GLOBUS_SUCCESS )
	    {
		globus_libc_free(dn);
		globus_libc_free(size_values[0]);
		globus_mutex_unlock(&handle->mutex);
		return GLOBUS_REPLICA_CATALOG_ERROR(
		    globus_error_get(err),
		    "Trying to create parent" );
	    }

	    /* Try to create the logicalfile entry again. */ 
	    if ( (lderr = ldap_add_s(handle->ld, dn, mods))
		!= LDAP_SUCCESS )
	    {
		globus_libc_free(dn);
		globus_libc_free(size_values[0]);
		globus_mutex_unlock(&handle->mutex);
		return GLOBUS_REPLICA_CATALOG_ERROR(
		    GLOBUS_NULL,
		    ldap_err2string(lderr) );
	    }
	}

	globus_libc_free(dn);
	globus_libc_free(size_values[0]);

	if ( lderr != LDAP_SUCCESS )
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_logicalfile_add_attribute(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          logicalfile_name,
    char *                                          attribute_name,
    char *                                          attribute_value )
{
    char *                              dn;
    char *                              attribute_values[2];
    LDAPMod                             attribute_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    attribute_values[0] = attribute_value;
    attribute_values[1] = NULL;
    attribute_mod.mod_op = LDAP_MOD_ADD;
    attribute_mod.mod_type = attribute_name;
    attribute_mod.mod_values = attribute_values;
    mods[0] = &attribute_mod;
    mods[1] = NULL;

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_logicalfile_dn(handle, logicalfile_name);
	
	lderr = ldap_modify_s(handle->ld, dn, mods);
    }
    globus_mutex_unlock(&handle->mutex);
    
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_logicalfile_delete(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          logicalfile_name )
{
    char *                              dn;
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_logicalfile_dn(handle, logicalfile_name);
	lderr = ldap_delete_s(handle->ld, dn);
    }
    globus_mutex_unlock(&handle->mutex);
    
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_logicalfile_delete_attribute(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          logicalfile_name,
    char *                                          attribute_name,
    char *                                          attribute_value )
{
    char *                              dn;
    char *                              attribute_values[2];
    LDAPMod                             attribute_mod;
    LDAPMod *                           mods[2];
    int                                 lderr;

    attribute_values[0] = attribute_value;
    attribute_values[1] = NULL;
    attribute_mod.mod_op = LDAP_MOD_DELETE;
    attribute_mod.mod_type = attribute_name;
    attribute_mod.mod_values = (attribute_value == GLOBUS_NULL)
	? NULL : attribute_values;
    mods[0] = &attribute_mod;
    mods[1] = NULL;

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_logicalfile_dn(handle, logicalfile_name);
	lderr = ldap_modify_s(handle->ld, dn, mods);
    }
    globus_mutex_unlock(&handle->mutex);
	
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_collection_list_logicalfiles(
    globus_replica_catalog_collection_handle_t *    handle,
    char **                                         attribute_names,
    struct globus_replica_catalog_entry_set_s *     logicalfiles )
{
    char *                              dn;
    char *                              filter = "(objectclass="
	GLOBUS_REPLICA_CATALOG_LOGICALFILE_OBJECTCLASS ")";
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
    
	dn = globus_l_replica_catalog_fileparent_dn(handle);

	lderr = ldap_search_s(
	    handle->ld,
	    dn,
	    LDAP_SCOPE_ONELEVEL,
	    filter,
	    (attribute_names != GLOBUS_NULL) ? attribute_names : NULL,
	    0,
	    &(logicalfiles->result) );
    }
    globus_mutex_unlock(&handle->mutex);
    
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    logicalfiles->collection = handle;

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_catalog_logicalfile_list_attributes(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          logicalfile_name,
    char **                                         attribute_names,
    struct globus_replica_catalog_attribute_set_s * attributes )
{
    char *                              dn;
    char *                              filter = "(objectclass=*)";
    LDAPMessage *                       result;
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
    
	dn = globus_l_replica_catalog_logicalfile_dn(handle, logicalfile_name);
	
	lderr = ldap_search_s(
	    handle->ld,
	    dn,
	    LDAP_SCOPE_BASE,
	    filter,
	    (attribute_names != GLOBUS_NULL) ? attribute_names : NULL,
	    0,
	    &result );

	globus_libc_free(dn);

	if (lderr != LDAP_SUCCESS)
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}

	attributes->collection = handle;
	attributes->result = result;
	attributes->entry = ldap_first_entry(handle->ld, result);
	
	ldap_get_option(handle->ld, LDAP_OPT_ERROR_NUMBER, &lderr);
	if (lderr != LDAP_SUCCESS)
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		GLOBUS_NULL,
		ldap_err2string(lderr) );
	}
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

/* Rollback and Restart Information Management */

globus_result_t
globus_replica_catalog_list_rollback_entries(
    char *                                           url,
    globus_replica_catalog_collection_handleattr_t * handleattrs,
    char **                                          attribute_names,
    struct globus_replica_catalog_entry_set_s *      rollback_entries )
{
    globus_result_t * err;
    globus_replica_catalog_collection_handleattr_t * attrs;
    globus_replica_catalog_collection_handle_t * handle;
    char * filter = "(objectclass="
	    GLOBUS_REPLICA_CATALOG_ROLLBACK_INFO_OBJECTCLASS ")";
    int lderr;
    
    attrs = (globus_replica_catalog_collection_handleattr_t *)
	    globus_libc_malloc(
		sizeof(globus_replica_catalog_collection_handleattr_t) );

    handle = (globus_replica_catalog_collection_handle_t *)
	    globus_libc_malloc(
		sizeof(globus_replica_catalog_collection_handle_t) );

    if (handleattrs == GLOBUS_NULL) {
	globus_replica_catalog_collection_handleattr_init(attrs);
    } else {
	globus_replica_catalog_collection_handleattr_copy(attrs,
		handleattrs);
    }

    if ( (err = globus_replica_catalog_collection_open(
	    handle,
	    attrs,
	    url ) ) != GLOBUS_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
		globus_error_get(err),
		"Can't connect to replica catalog" );
    }

    lderr = ldap_search_s(
	handle->ld,
	handle->dn,
	LDAP_SCOPE_SUBTREE,
	filter,
	(attribute_names != GLOBUS_NULL) ? attribute_names : NULL,
	0,
	&(rollback_entries->result) );
    
    if ( lderr != LDAP_SUCCESS )
    {
	globus_replica_catalog_collection_close(handle);
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    /* the open above is closed in globus_replica_catalogentry_set_destroy */
    rollback_entries->collection = handle;
    rollback_entries->collection_close = GLOBUS_TRUE;

    return GLOBUS_SUCCESS;
}

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL

/**
 * Construct the DN for the rollback and restart information for a
 * logical file in a location name.
 *
 * Result must be freed by the caller with globus_libc_free().
 *
 * @param handle
 * 	  handle to the collection
 * @param location_name
 * 	  location name of the target
 * @param logicalfile_name
 * 	  logical file name of the target
 */
static
char *
globus_l_replica_catalog_rollback_info_dn(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char *                                          logicalfile_name )
{
    char *                              location_dn;
    char *                              dn;

    location_dn = globus_l_replica_catalog_location_dn( handle,
	    location_name );

    dn = (char *) globus_libc_malloc(
	strlen(GLOBUS_REPLICA_CATALOG_ROLLBACK_INFO_RDN_TYPE)
	+ strlen(logicalfile_name)
	+ strlen(location_dn)
	+ 4 );

    sprintf( dn, "%s=%s, %s",
	GLOBUS_REPLICA_CATALOG_ROLLBACK_INFO_RDN_TYPE,
	logicalfile_name,
	location_dn ); 

    globus_libc_free(location_dn);

    return dn;
}

#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

globus_result_t
globus_i_replica_catalog_rollback_info_construct_contact(
    globus_replica_catalog_collection_handle_t *    handle,
    char *                                          location_name,
    char *                                          logicalfile_name,
    char **                                         contact_string )
{
    char *                              dn;

    dn = globus_l_replica_catalog_rollback_info_dn( handle,
	    location_name, logicalfile_name );

    *contact_string = (char *) globus_libc_malloc(
	strlen(handle->urld->lud_host)
	+ 5 + strlen(dn) + 10 );

    sprintf( *contact_string, "ldap://%s:%d/%s",
	handle->urld->lud_host,
	handle->urld->lud_port,
	dn );

    globus_libc_free(dn);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_i_replica_catalog_rollback_info_parse_contact(
    char *                                          contact_string,
    char **                                         url,
    char **                                         location_name,
    char **                                         logicalfile_name )
{
    LDAPURLDesc *			urld;
    int					ldurlerr;
    char *				ldurlerrstr;
    int					lderr;
    char **				rdns;
    char **				names;
    int					i;
    int					len;
    char *				collection_dn;

    ldurlerr = ldap_url_parse(contact_string, &urld);
    /* FIXME: Solaris is choking on this check, don't know why. */
    /* removed the if 0 around this code, seems to work on solaris */

    if (ldurlerr != 0) 
    {
	sprintf(ldurlerrstr, "Error parsing contact string %s",
	    contact_string);
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldurlerrstr );
    }

    names = ldap_explode_dn(urld->lud_dn, 1);
    if (names != NULL) {
	*logicalfile_name = globus_libc_strdup(names[0]);
	*location_name = globus_libc_strdup(names[1]);
    } else {
	*logicalfile_name = GLOBUS_NULL;
	*location_name = GLOBUS_NULL;
    }
    ldap_value_free(names);

    rdns = ldap_explode_dn(urld->lud_dn, 0);
    len = 1;
    for (i = 2; rdns[i] != NULL; ++i) {
	len += 2 + strlen(rdns[i]);
    }
    collection_dn = (char *) globus_libc_malloc(len);
    *collection_dn = '\0';
    for (i = 2; rdns[i] != NULL; ++i) {
	if (i > 2) {
	    strcat(collection_dn, ", ");
	}
	strcat(collection_dn, rdns[i]);
    }
    ldap_value_free(rdns);

    *url = (char *) globus_libc_malloc( 14 + strlen(urld->lud_host)
	    + strlen(collection_dn) );
    if (urld->lud_port == 0) {
	sprintf(*url, "ldap://%s/%s", urld->lud_host, collection_dn);
    } else {
	sprintf(*url, "ldap://%s:%d/%s", urld->lud_host, urld->lud_port,
		collection_dn);
    }
    globus_libc_free(collection_dn);

    ldap_free_urldesc(urld);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_i_replica_catalog_rollback_info_create(
    globus_replica_catalog_collection_handle_t *	handle,
    char *						location_name,
    char *						logicalfile_name,
    char *						operation,
    char *						state,
    char *						session_id,
    char *						source_url )
{
    char *                              dn;
    char *                              objectclass_values[4];
    char *                              operation_values[2];
    char *                              state_values[2];
    char *                              source_values[2];
    char *                              sessionid_values[2];
    LDAPMod                             objectclass_mod;
    LDAPMod                             operation_mod;
    LDAPMod                             state_mod;
    LDAPMod                             source_mod;
    LDAPMod                             sessionid_mod;
    LDAPMod *                           mods[6];
    int                                 lderr;

    objectclass_values[0] = "top";
    objectclass_values[1] = "GlobusTop";
    objectclass_values[2] = GLOBUS_REPLICA_CATALOG_ROLLBACK_INFO_OBJECTCLASS;
    objectclass_values[3] = NULL;
    objectclass_mod.mod_op = 0;
    objectclass_mod.mod_type = "objectclass";
    objectclass_mod.mod_values = objectclass_values;

    operation_values[0] = operation;
    operation_values[1] = NULL; 
    operation_mod.mod_op = 0;
    operation_mod.mod_type = "operation";
    operation_mod.mod_values = operation_values;

    state_values[0] = state;
    state_values[1] = NULL; 
    state_mod.mod_op = 0;
    state_mod.mod_type = "state";
    state_mod.mod_values = state_values;

    if (source_url != GLOBUS_NULL) {
	source_values[0] = source_url;
	source_values[1] = NULL; 
	source_mod.mod_op = 0;
	source_mod.mod_type = "source";
	source_mod.mod_values = source_values;
    }

    if (session_id != GLOBUS_NULL) {
	sessionid_values[0] = session_id;
	sessionid_values[1] = NULL; 
	sessionid_mod.mod_op = 0;
	sessionid_mod.mod_type = "sessionid";
	sessionid_mod.mod_values = sessionid_values;
    }

    mods[0] = &objectclass_mod;
    mods[1] = &operation_mod;
    mods[2] = &state_mod;
    mods[3] = NULL ;
    mods[4] = NULL ;
    mods[5] = NULL;
    if (source_url != GLOBUS_NULL) {
	mods[3] = &source_mod;
	if (session_id != GLOBUS_NULL) {
	    mods[4] = &sessionid_mod;
	}
    } else if (session_id != GLOBUS_NULL) {
	mods[3] = &sessionid_mod;
    }

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_rollback_info_dn( handle,
		location_name, logicalfile_name );

	lderr = ldap_add_s(handle->ld, dn, mods);
    }
    globus_mutex_unlock(&handle->mutex);
	
    globus_libc_free(dn);
	
    if ( lderr != LDAP_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_i_replica_catalog_rollback_info_update(
    globus_replica_catalog_collection_handle_t *	handle,
    char *						location_name,
    char *						logicalfile_name,
    char *						state,
    char *						ranges )
{
    char *                              state_values[2];
    char *                              ranges_values[2];
    LDAPMod                             state_mod;
    LDAPMod                             ranges_mod;
    LDAPMod *                           mods[3];
    char *                              dn;
    int                                 lderr;

    if (state != GLOBUS_NULL) {
	state_values[0] = state;
	state_values[1] = NULL; 
	state_mod.mod_op = LDAP_MOD_REPLACE;
	state_mod.mod_type = "state";
	state_mod.mod_values = state_values;
    }

    if (ranges != GLOBUS_NULL) {
	ranges_values[0] = ranges;
	ranges_values[1] = NULL; 
	ranges_mod.mod_op = LDAP_MOD_REPLACE;
	ranges_mod.mod_type = "ranges";
	ranges_mod.mod_values = ranges_values;
    }

    mods[0] = NULL;
    mods[1] = NULL;
    mods[2] = NULL;
    if (state != GLOBUS_NULL) {
	mods[0] = &state_mod;
	if (ranges != GLOBUS_NULL) {
	    mods[1] = &ranges_mod;
	}
    } else if (ranges != GLOBUS_NULL) {
	mods[0] = &ranges_mod;
    }

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_rollback_info_dn( handle,
		location_name, logicalfile_name );

	lderr = ldap_modify_s(handle->ld, dn, mods);
    }
    globus_mutex_unlock(&handle->mutex);
	
    globus_libc_free(dn);
	
    if ( lderr != LDAP_SUCCESS )
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_i_replica_catalog_rollback_info_get(
    globus_replica_catalog_collection_handle_t *	handle,
    char *						location_name,
    char *						logicalfile_name,
    char **						operation,
    char **						state,
    char **						session_id,
    char **						source_url,
    char **						ranges )
{
    char *                              dn;
    char *                              filter = "(objectclass=*)";
    LDAPMessage *                       result;
    LDAPMessage *                       entry;
    int                                 lderr;
    char **				values;

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_rollback_info_dn( handle,
						       location_name,
						       logicalfile_name );

	lderr = ldap_search_s( handle->ld,
			       dn, LDAP_SCOPE_BASE,
			       filter, NULL, 0,
			       &result );

	globus_libc_free(dn);

	if (lderr != LDAP_SUCCESS)
	{
	    globus_mutex_unlock(&handle->mutex);
	    return GLOBUS_REPLICA_CATALOG_ERROR(
		    GLOBUS_NULL,
		    ldap_err2string(lderr) );
	}

	if ( ( entry = ldap_first_entry( handle->ld,
					 result ) ) != NULL )
	{
	    if ( (values = ldap_get_values( handle->ld, entry,
		    "operation" ) ) == NULL )
	    {
		*operation = GLOBUS_NULL;
		ldap_msgfree(result);
		globus_mutex_unlock(&handle->mutex);
		return GLOBUS_REPLICA_CATALOG_ERROR(
			GLOBUS_NULL,
			"No operation information" );
	    } else {
		*operation = globus_libc_strdup(values[0]);
		ldap_value_free(values);
	    }

	    if ( (values = ldap_get_values( handle->ld, entry,
		    "state" ) ) == NULL )
	    {
		*state = GLOBUS_NULL;
		ldap_msgfree(result);
		globus_mutex_unlock(&handle->mutex);
		return GLOBUS_REPLICA_CATALOG_ERROR(
			GLOBUS_NULL,
			"No state information" );
	    } else {
		*state = globus_libc_strdup(values[0]);
		ldap_value_free(values);
	    }

	    if ( (values = ldap_get_values( handle->ld, entry,
		    "sessionid" ) ) == NULL )
	    {
		*session_id = GLOBUS_NULL;
	    } else {
		*session_id = globus_libc_strdup(values[0]);
		ldap_value_free(values);
	    }

	    if ( (values = ldap_get_values( handle->ld, entry,
		    "source" ) ) == NULL )
	    {
		*source_url = GLOBUS_NULL;
	    } else {
		*source_url = globus_libc_strdup(values[0]);
		ldap_value_free(values);
	    }

	    if ( (values = ldap_get_values( handle->ld, entry,
		    "ranges" ) ) == NULL )
	    {
		*ranges = GLOBUS_NULL;
	    } else {
		*ranges = globus_libc_strdup(values[0]);
		ldap_value_free(values);
	    }
	} else {
	    ldap_get_option(handle->ld, LDAP_OPT_ERROR_NUMBER, &lderr);
	    if (lderr != LDAP_SUCCESS)
	    {
		globus_mutex_unlock(&handle->mutex);
		return GLOBUS_REPLICA_CATALOG_ERROR(
		    GLOBUS_NULL,
		    ldap_err2string(lderr) );
	    }
	}
	
	ldap_msgfree(result);
    }
    globus_mutex_unlock(&handle->mutex);

    return GLOBUS_SUCCESS;
}

globus_result_t
globus_i_replica_catalog_rollback_info_delete(
    globus_replica_catalog_collection_handle_t *	handle,
    char *						location_name,
    char *						logicalfile_name )
{
    char *                              dn;
    int                                 lderr;

    globus_mutex_lock(&handle->mutex);
    {
	dn = globus_l_replica_catalog_rollback_info_dn( handle,
		location_name,
		logicalfile_name );
	
	lderr = ldap_delete_s(handle->ld, dn);
    }
    globus_mutex_unlock(&handle->mutex);
	
    globus_libc_free(dn);

    if (lderr != LDAP_SUCCESS)
    {
	return GLOBUS_REPLICA_CATALOG_ERROR(
	    GLOBUS_NULL,
	    ldap_err2string(lderr) );
    }

    return GLOBUS_SUCCESS;
}

